/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/* 
 * elem_play.c --
 *
 *      This file contains C code that implements a random-access decoding.
 *
 */

#include "global.h"
#include "elem_play.h"
#include "video.h"
#include "readfile.h"
#include "util.h"


/* Declarations of functions. */
static void DoPictureDisplay();


/*
 *--------------------------------------------------------------
 *
 * mpegVidParse --
 *
 *      Parses bit stream by frames, gops, and other major
 *	markers for division into pointers into the file.
 *
 * Results:
 *      The kind of element just parsed.
 *
 * Side effects:
 *      Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

BOOLEAN
mpegVidParse(data_stream, callback, ref)
  DataStream *data_stream;
  ParseCallback *callback;
  void *ref;
{
  unsigned int data;
  Sequence seq;
  GoP group;
  Pict pict;
  long fpos;
  BOOLEAN ret = TRUE;
  BOOLEAN first = TRUE;

  if (data_stream == NULL || callback == NULL)
    return FALSE;

  next_start_code(data_stream);  /* sets curBits */
  show_bits32(data);
  if (data != SEQ_START_CODE)
    return FALSE;

  InitSequence(&seq);
  InitGoP(&group);
  InitPict(&pict);

  for (;;) {
    /* Save current position */
    fpos = TellStream(data_stream);

    /* Get next 32 bits (size of start codes). */
    show_bits32(data);

    /*
     * Process according to start code.
     */

    switch (data) {

    case SEQ_END_CODE:
    case ISO_11172_END_CODE:

      /* Sequence done. Report and exit. */

      if (!callback(fpos, VID_END, NULL, ref))
	goto error;
      goto done;

    case SEQ_START_CODE:

      if (first) {
	first = FALSE;

	/* Sequence start code. Parse sequence header. */

	if (ParseSequence(data_stream, &seq) != PARSE_OK)
	  goto error;
	
	if (!callback(fpos, VID_START, &seq, ref))
	  goto error;
      }
      else {
	fprintf(stderr, "Warning: extraneous start code in video stream.\n");

	flush_bits(24);
      }
      break;

    case GOP_START_CODE:

      /* Group of Pictures start code. Parse group header. */

      if (ParseGoP(data_stream, &group) != PARSE_OK)
	goto error;

      if (!callback(fpos, VID_GOP, &group, ref))
	goto error;
      break;

    case PICTURE_START_CODE:

      /* Picture start code. Parse picture header. */

      if (ParsePicture(data_stream, &pict) != PARSE_OK)
	goto error;

      if (!callback(fpos, VID_FRAME, &pict, ref))
	goto error;

      break;

    case SEQUENCE_ERROR_CODE:

      goto error;

    default:
      flush_bits(24);
      break;
    }

    next_start_code(data_stream);
  }

error:
  ret = FALSE;

done:
  DestructPict(&pict);
  DestructGoP(&group);
  DestructSequence(&seq);

  return ret;
}




/*
 *--------------------------------------------------------------
 *
 * mpegVidStart --
 *
 *      Parses and loads the start of the video stream.
 *
 * Results:
 *      A value indicating whether the movie is OK.
 *
 * Side effects:
 *      Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

BOOLEAN
mpegVidStart(data_stream, vid_decoder, xinfo)
  DataStream *data_stream;
  VidDecoder *vid_decoder;
  XInfo     *xinfo;
{
  unsigned int data;

  if (data_stream == NULL || vid_decoder == NULL)
    return FALSE;

  /*
   * Find start code and make sure it is a
   * sequence start code.
   */

  next_start_code(data_stream);  /* sets curBits */
  show_bits32(data);
  if (data != SEQ_START_CODE)
    return FALSE;

  /* Sequence start code. Parse sequence header. */

  if (ParseSequence(data_stream, &vid_decoder->sequence) != PARSE_OK)
    return FALSE;

  StartVidSeq(vid_decoder, xinfo);
  return TRUE;
}




/*
 *--------------------------------------------------------------
 *
 * mpegVidGoP --
 *
 *      Parses out a GoP for the decoding sequence.
 *
 * Results:
 *      A value indicating whether the movie is OK.
 *
 * Side effects:
 *      Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

BOOLEAN
mpegVidGoP(data_stream, vid_decoder)
  DataStream *data_stream;
  VidDecoder *vid_decoder;
{
  unsigned int data;

  if (data_stream == NULL || vid_decoder == NULL)
    return FALSE;

  /* Get next 32 bits (size of start codes). */

  show_bits32(data);

  if (data != GOP_START_CODE)
    return FALSE;

  /* Group of Pictures start code. Parse group header. */

  if (ParseGoP(data_stream, &vid_decoder->group) != PARSE_OK)
    return FALSE;
  
  return TRUE;
}




/*
 *--------------------------------------------------------------
 *
 * mpegVidFrame --
 *
 *	Parse out a frame of video from the stream.
 *
 * Results:
 *      A value indicating whether the movie is OK.
 *
 * Side effects:
 *      Bit stream is irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

BOOLEAN
mpegVidFrame(data_stream, vid_decoder, destImage, pastImage, futureImage)
  DataStream *data_stream;
  VidDecoder *vid_decoder;
  PictImage *destImage;
  PictImage *pastImage;
  PictImage *futureImage;
{
  unsigned int data;
  int status;

  if (data_stream == NULL || vid_decoder == NULL)
    return FALSE;

  for (;;) {
    /* Get next 32 bits (size of start codes). */

    show_bits32(data);

    /*
     * Process according to start code (or parse macroblock if not a
     * start code at all).
     */

    switch (data) {

    case SEQ_START_CODE:
    case SEQ_END_CODE:
    case ISO_11172_END_CODE:
    case GOP_START_CODE:

      /* We shouldn't have gotten one of these... */
      return FALSE;

    case PICTURE_START_CODE:

      /* Picture start code. Parse picture header and first slice header. */

      status = ParsePicture(data_stream, &vid_decoder->picture);

      if (status == PARSE_OK)
	status = StartPicture(vid_decoder, destImage, pastImage, futureImage);
	  
      if (status == SKIP_PICTURE)
	return TRUE;
      else if (status != PARSE_OK)
	return FALSE;

      if (ParseSlice(data_stream, &vid_decoder->slice) != PARSE_OK)
	return FALSE;

      StartSlice(vid_decoder);
      goto slice;

    case SEQUENCE_ERROR_CODE:
      flush_bits32;
      next_start_code(data_stream);
      return FALSE;
      
    default:

      /* Check for slice start code. */

      if ((data >= SLICE_MIN_START_CODE) && (data <= SLICE_MAX_START_CODE)) {

	/* Slice start code. Parse slice header. */

	if (ParseSlice(data_stream, &vid_decoder->slice) != PARSE_OK)
	  return FALSE;

	StartSlice(vid_decoder);
      }
      goto slice;
    }
    continue;

  slice:
    /* Parse macroblocks. */

    for (;;) {
      /* Check to see if actually a startcode and not a macroblock. */

      if (next_bits(23, 0x00000000, data_stream))
	goto end_mb;

      /* Not start code. Parse Macroblock. */

      if (ParseDecodeMB(data_stream, vid_decoder) != PARSE_OK)
	return FALSE;
    }	
    continue;

  end_mb:
    /* Not macroblock, actually start code. Get start code. */

    next_start_code(data_stream);
    show_bits32(data);

    /*
     * If start code is outside range of slice start codes, frame is
     * complete, dither frame.
     */

    if (((data < SLICE_MIN_START_CODE) || (data > SLICE_MAX_START_CODE)) &&
	(data != SEQUENCE_ERROR_CODE)) {
      
      DoDitherImage(vid_decoder);
      return TRUE;
    }
    continue;
  }
}


