/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/*
 * readfile.c --
 *
 *       Procedures concerned with reading data and parsing 
 *       start codes from MPEG files.
 *
 */

#include "global.h"

#include <assert.h>
#include <sys/types.h>
#include <signal.h>
#ifndef MIPS
#include <netinet/in.h>
#else
#include <bsd/netinet/in.h>
#endif

#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#endif

#include "readfile.h"

#ifdef HAVE_LIBSWQ
#include <swq_smba.h>
#endif

#ifdef SYS_STREAM
#include "video.h"
#include "util.h"
#include "dither.h"


/* Silly Constants.... */
#define PACK_START_CODE             0x000001ba
#define SYSTEM_HEADER_START_CODE    0x000001bb
#define PACKET_START_CODE_MASK      0xffffff00
#define PACKET_START_CODE_PREFIX    0x00000100

#define PACK_HEADER_SIZE 8

#define STD_AUDIO_STREAM_ID ((unsigned char) 0xb8)
#define STD_VIDEO_STREAM_ID ((unsigned char) 0xb9)
#define MIN_STREAM_ID_ID    ((unsigned char) 0xbc)
#define RESERVED_STREAM_ID  ((unsigned char) 0xbc)
#define PRIVATE_STREAM_1_ID ((unsigned char) 0xbd)
#define PADDING_STREAM_ID   ((unsigned char) 0xbe)
#define PRIVATE_STREAM_2_ID ((unsigned char) 0xbf)

#define STD_SYSTEM_CLOCK_FREQ 90000ul
#define MUX_RATE_SCALE_FACTOR 50
#define MAX_STREAMS 8
#define NOT_PACKET_ID       ((unsigned char) 0xff)
#define KILL_BUFFER         ((unsigned char) 0xfe)

#endif SYS_STREAM

/* Bit masks used by bit i/o operations. */

unsigned int nBitMask[] = { 0x00000000, 0x80000000, 0xc0000000, 0xe0000000, 
  0xf0000000, 0xf8000000, 0xfc000000, 0xfe000000, 
  0xff000000, 0xff800000, 0xffc00000, 0xffe00000, 
  0xfff00000, 0xfff80000, 0xfffc0000, 0xfffe0000, 
  0xffff0000, 0xffff8000, 0xffffc000, 0xffffe000, 
  0xfffff000, 0xfffff800, 0xfffffc00, 0xfffffe00, 
  0xffffff00, 0xffffff80, 0xffffffc0, 0xffffffe0, 
  0xfffffff0, 0xfffffff8, 0xfffffffc, 0xfffffffe};

  unsigned int bitMask[] = {  0xffffffff, 0x7fffffff, 0x3fffffff, 0x1fffffff, 
    0x0fffffff, 0x07ffffff, 0x03ffffff, 0x01ffffff,
    0x00ffffff, 0x007fffff, 0x003fffff, 0x001fffff,
    0x000fffff, 0x0007ffff, 0x0003ffff, 0x0001ffff,
    0x0000ffff, 0x00007fff, 0x00003fff, 0x00001fff,
    0x00000fff, 0x000007ff, 0x000003ff, 0x000001ff,
    0x000000ff, 0x0000007f, 0x0000003f, 0x0000001f,
    0x0000000f, 0x00000007, 0x00000003, 0x00000001};

    unsigned int rBitMask[] = { 0xffffffff, 0xfffffffe, 0xfffffffc, 0xfffffff8, 
      0xfffffff0, 0xffffffe0, 0xffffffc0, 0xffffff80, 
      0xffffff00, 0xfffffe00, 0xfffffc00, 0xfffff800, 
      0xfffff000, 0xffffe000, 0xffffc000, 0xffff8000, 
      0xffff0000, 0xfffe0000, 0xfffc0000, 0xfff80000, 
      0xfff00000, 0xffe00000, 0xffc00000, 0xff800000, 
      0xff000000, 0xfe000000, 0xfc000000, 0xf8000000, 
      0xf0000000, 0xe0000000, 0xc0000000, 0x80000000};

      unsigned int bitTest[] = {  0x80000000, 0x40000000, 0x20000000, 0x10000000, 
	0x08000000, 0x04000000, 0x02000000, 0x01000000,
	0x00800000, 0x00400000, 0x00200000, 0x00100000,
	0x00080000, 0x00040000, 0x00020000, 0x00010000,
	0x00008000, 0x00004000, 0x00002000, 0x00001000,
	0x00000800, 0x00000400, 0x00000200, 0x00000100,
	0x00000080, 0x00000040, 0x00000020, 0x00000010,
	0x00000008, 0x00000004, 0x00000002, 0x00000001};

	
	/*
	 *--------------------------------------------------------------
	 *
	 * NewDataStream --
	 *
	 *	Allocates and initializes a DataStream structure. Takes
	 *      as parameter requested size for buffer length.
 *
 * Results:
 *	A pointer to the new DataStream structure.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

DataStream *
NewDataStream(buffer_len)
  unsigned int buffer_len;
{
  DataStream *new;
  DBUG_ENTER();

  /* Check for legal buffer length. */

  if (buffer_len < 0)
    DBUG_RETURN(NULL);

  /* Make buffer length multiple of 4. */

  buffer_len = (buffer_len + 3) >> 2;

  /* Allocate memory for new structure. */

  new = (DataStream *) malloc(sizeof(DataStream));

  /* Create buffer. */
  
  new->buf_start = (unsigned int *) malloc(buffer_len * 4);

  /*
   * Set max_buf_length to one less than actual length to deal with messy
   * data without proper seq. end codes.
   */

  new->max_buf_length = buffer_len - 1;

  /* Initialize bitstream i/o fields... */

  new->bit_offset = 0;
  new->buf_length = 0;
  new->buffer = new->buf_start;

  /* And other fields. */

  new->input = NULL;
  new->EOF_flag = FALSE;
  new->EOF_adjust = 0;

#ifdef HAVE_LIBSWQ
  /* SWQ fields */
  new->queue = 0;
  new->buf_area = 0;
#endif

#ifdef SYS_STREAM
  /* System stream stuff... */
  new->sys_layer=-1;
  new->num_left=0;
  new->leftover_bytes=0;
#endif

#if HAVE_CHANNELS
  new->channel_EOF=0;
#endif

  /* Return structure. */

  DBUG_RETURN(new);
}



/*
 *--------------------------------------------------------------
 *
 * DestroyVidDecoder --
 *
 *	Deallocates a VidDecoder structure.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

void
DestroyDataStream(data)
  DataStream *data;
{
  DBUG_ENTER();
#ifdef HAVE_LIBSWQ
  if (data->buf_area)
    smbaclose(data->buf_area);
  if (data->queue)
    swqClose(data->queue);
#endif

  if (data->buf_start)
    free(data->buf_start);

  free((char *) data);
  DBUG_VOID_RETURN();
}


/*
 *-------------------------------------------------------------
 *
 * TellStream
 *
 * Determine the current offset into a DataStream.
 *
 *-------------------------------------------------------------
 */
long
TellStream(data_stream)
  DataStream *data_stream;
{
  long pos;
  DBUG_ENTER();

  assert(data_stream->input);
  assert((data_stream->bit_offset & 7) == 0);
  
  pos = ftell(data_stream->input);
  if (pos < 0)
    abort();

  pos -= data_stream->buf_length << 2;
  pos += data_stream->bit_offset >> 3;
  pos += data_stream->EOF_adjust;

  DBUG_RETURN(pos);
}



/*
 *-------------------------------------------------------------
 *
 * SeekStream
 *
 * Goto an offset in the steam
 *
 *-------------------------------------------------------------
 */
void
SeekStream(data_stream, seekValue)
  DataStream *data_stream;
  long seekValue;
{
  int errno;
  int fpos;
  DBUG_ENTER();
  
  fpos = TellStream(data_stream);
  if (seekValue == fpos)
    DBUG_VOID_RETURN();	/* XXX Opt if near? */

#ifdef SEEK_SET
  errno = fseek(data_stream->input, seekValue, SEEK_SET);
#else
  errno = fseek(data_stream->input, seekValue, 0);
#endif
  if (errno != 0)
    abort();

  /* clear that buffer */
  data_stream->buffer = data_stream->buf_start;
  data_stream->buf_length = 0;
  data_stream->bit_offset = 0;
  data_stream->EOF_flag = 0;
  data_stream->EOF_adjust = 0;
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * correct_underflow --
 *
 *	Called when buffer does not have sufficient data to 
 *      satisfy request for bits.
 *      Calls get_more_data, an application specific routine
 *      required to fill the buffer with more data.
 *
 * Results:
 *      None really.
 *  
 * Side effects:
 *	buf_length and buffer fields may be changed.
 *
 *--------------------------------------------------------------
 */

void 
correct_underflow(data_stream)
   DataStream *data_stream;
{

  int status;
  DBUG_ENTER();

  status = get_more_data(data_stream);

  if (status  < 0) {
    exit(1);
  }
  else if ((status == 0) && (data_stream->buf_length < 1)) {

    DBUG_VOID_RETURN();
  }
  data_stream->curBits = *data_stream->buffer << data_stream->bit_offset;
  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * next_bits --
 *
 *	Compares next num bits to low order position in mask.
 *      Buffer pointer is NOT advanced.
 *
 * Results:
 *	TRUE, FALSE, or error code.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

int next_bits(num, mask, data_stream)
int num;
unsigned int mask;
DataStream *data_stream;
{
  unsigned int stream;
  int ret_value;
  DBUG_ENTER();

  assert(data_stream);

  /* Get next num bits, no buffer pointer advance. */

  show_bitsn(num, stream);

  /* Compare bit stream and mask. Set return value to TRUE if equal, FALSE if
     differs. 
  */

  if (mask == stream) {
    ret_value = TRUE;
  } else ret_value = FALSE;

  /* Return return value. */
  DBUG_RETURN(ret_value);
}


/*
 *--------------------------------------------------------------
 *
 * back_bits --
 *
 *      Unreads the last n bits read.  Only works before the latest
 *	call to correct_underflow().
 *
 * Results:
 *	TRUE or FALSE.  (TRUE if it worked.)
 *
 * Side effects:
 *	Stream wasn't parsed quite so irreversably after all.
 *
 *--------------------------------------------------------------
 */

int back_bits(num, data_stream)
int num;
DataStream *data_stream;
{
  /* Set buffer pointer back and reset length & bit offsets so
   * next bytes will be beginning of start code. 
   */

  int offset = data_stream-> bit_offset - num;
  unsigned int *buffer = data_stream->buffer;
  int length = data_stream->buf_length;
  DBUG_ENTER();

  while (offset < 0) {
    offset += 32;
    length++;
    buffer--;
  }
  if (buffer < data_stream->buf_start)
    DBUG_RETURN(FALSE);

  data_stream->buffer = buffer;
  data_stream->buf_length = length;
  data_stream->bit_offset = offset;
  data_stream->curBits = *buffer << offset;

  DBUG_RETURN(TRUE);
}


/*
 *--------------------------------------------------------------
 *
 * get_more_data --
 *
 *	Called by get_more_data to read in more data from
 *      video MPG files (non-system-layer)
 *
 * Results:
 *	Input buffer updated, buffer length updated.
 *      Returns 1 if data read, 0 if EOF, -1 if error.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */
int 
get_more_data(data_stream)
      DataStream *data_stream;
{
  unsigned int **bs_ptr=&data_stream->buf_start; 
  int *max_length=&data_stream->max_buf_length;
  int *length_ptr=&data_stream->buf_length;
  unsigned int **buf_ptr=&data_stream->buffer;
  int ioBytes, data, result;
  unsigned int *mark;
  DBUG_ENTER();

#ifdef SYS_STREAM
  int sys_layer= data_stream->sys_layer;
  
  if (sys_layer == 0) {
#endif
    DBUG_RETURN(pure_get_more_data(bs_ptr,
			      max_length,
			       length_ptr,
			       buf_ptr,
			       data_stream));
#ifdef SYS_STREAM
  }

  if (sys_layer == -1) {
    /* Time to init ourselves */
    data_stream->swap = (htonl(0x1234) != 0x1234);
    mark = *bs_ptr;
    ioBytes = fread(&data, 1, 4, data_stream->input);

    if (ioBytes != 4) {
	  DBUG_RETURN(0);
    }

    data = ntohl(data);
    if ( (data == PACK_START_CODE) || (data == SYSTEM_HEADER_START_CODE) ) {
      /* Yow, a System Layer Stream.  Much harder to parse.  Call in the
	     specialist.... */
      fprintf(stderr,"This is an MPEG System Layer Stream.  ");
      fprintf(stderr,"Audio is not played.\n");
      data_stream->sys_layer = 1;
      result = read_sys(data_stream,(unsigned int) data);
      DBUG_RETURN(result);
    } else {
      /* No system Layer junk, just pretent we didn't peek,
	     and hereafter just call pure_get_more_data */
      data_stream->sys_layer = 0;
      **bs_ptr = data;
      *length_ptr = 1;
      result = pure_get_more_data(*bs_ptr, *max_length, 
				 length_ptr, buf_ptr, data_stream);
      *buf_ptr = *bs_ptr;
      DBUG_RETURN(result);
    }
  }

  /* A system layer stream (called after the 1st time), call the specialist */
  result = read_sys(data_stream,0);
  DBUG_RETURN(result);
#endif
}

#ifdef SYS_STREAM

/*
 *-------------------------------------------------------------
 *
 * clear_data_stream
 *
 * Empties out internal buffers
 *
 *-------------------------------------------------------------
 */
void 
  clear_data_stream(data_stream)
     DataStream *data_stream;
{
  DBUG_ENTER();
  /* Only internal buffer is in ReadPacket */
  if (data_stream->sys_layer) {
    ReadPacket(KILL_BUFFER, data_stream);
  }
  DBUG_VOID_RETURN();
}
#endif


/*
 *--------------------------------------------------------------
 *
 * pure_get_more_data --
 *      (get_more_data from ver 2.0 with swap added)
 *
 *	Called by get_more_data to read in more data from
 *      video MPG files (non-system-layer)
 *
 * Results:
 *	Input buffer updated, buffer length updated.
 *      Returns 1 if data read, 0 if EOF, -1 if error.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

int 
pure_get_more_data(bs_ptr, max_length_ptr, length_ptr, buf_ptr, data_stream)
     unsigned int **bs_ptr;
     int *max_length_ptr;
     int *length_ptr;
     unsigned int **buf_ptr;
     DataStream *data_stream;
{
  
  int length, num_read, i;
  unsigned int request;
  unsigned int *buf_start = *bs_ptr;
  int max_length = *max_length_ptr;
  unsigned char *mark;
  unsigned int *lmark;
  DBUG_ENTER();
  
  if (data_stream->EOF_flag) DBUG_RETURN(0);
  
  length = *length_ptr;
  
  if (length > 0) {
    memmove(buf_start, *buf_ptr, (unsigned int) (length << 2));
    mark = ((unsigned char *) (buf_start + length));
  }
  else {
    mark = (unsigned char *) buf_start;
    length = 0;
  }
  
  request = (max_length-length) << 2;
  
#if defined(HAVE_LIBSWQ) || defined(HAVE_CHANNELS)
  if (data_stream->input) {
#endif // HAVE_LIBSWQ || HAVE_CHANNELS

    /* XXX fread is silly here... although more portable. */
    num_read = fread(mark, 1, request, data_stream->input);

#ifdef HAVE_LIBSWQ
  }
  else {
    char *msg;

    assert(data_stream->queue && data_stream->buf_area);

    msg = swqRecvSMBA(data_stream->queue, data_stream->buf_area, 0);
    if (!msg) {
      perror("swqRecvSMBA");
      exit(1);
    }

    num_read = SMB_BUFSIZE(msg);
    if (request < num_read) {
      /* Grow buffer.  Sigh. */

      request = ((length << 2) + num_read + 7) & ~3;
      if (request < ((max_length + 1) << 1))
	  request = (max_length + 1) << 1;

      *bs_ptr = buf_start = realloc(buf_start, request);
      if (!buf_start)
	abort();

      *max_length_ptr = max_length = (request >> 2) - 1;
      mark = (unsigned char *) (buf_start + length);
    }

    memcpy(mark, msg, num_read);

    if (smbafree(data_stream->buf_area, msg) < 0) {
      perror("smbafree");
      exit(1);
    }
  }
#elif defined(HAVE_CHANNELS)
  } else {
    dbuf *packet;
    packet = (data_stream->channel_EOF ? NULL : chRECV(data_stream->channelin));
    if ((!packet) || (!dbLEN(packet))) {
      data_stream->channel_EOF = 1;
      num_read = 0;
    } else {
      num_read = dbLEN(packet);
      if (request < num_read) {
	/* Grow buffer.  Sigh. */
	request = ((length << 2) + num_read + 7) & ~3;
	if (request < ((max_length + 1) << 1))
	  request = (max_length + 1) << 1;

	*bs_ptr = buf_start = realloc(buf_start, request);
	if (!buf_start)
	  abort();

	*max_length_ptr = max_length = (request >> 2) - 1;
	mark = (unsigned char *) (buf_start + length);
      }
      memcpy(mark, dbCUR(packet), num_read);
      dbFREE(packet);
    }
  } // else
#endif // HAVE_CHANNELS
  
  /* Correction for 4-byte alignment */
  {
    unsigned char *index;
    unsigned char *end;
    
    if (num_read & 0x3) {
      int adjust;
      
      index = mark + num_read;
      adjust = 4 - (num_read & 0x3);
      end = index + adjust;

      /* fill in with zeros */
      for (; index < end; index++)
	*index = 0;

      num_read += adjust;
      data_stream->EOF_adjust = adjust;
    }
  }
  
  if (num_read < 0) {
    DBUG_RETURN(-1);
  } else if (num_read == 0) {
    *buf_ptr = buf_start;
    
    /* Make 32 bits after end equal to 0 and 32
     * bits after that equal to seq end code
     * in order to prevent messy data from infinite
     * recursion.
     */
    
    *(buf_start + length) = 0x0;
    *(buf_start + length+1) = 0x000001b7;	/* XXX (yuck) SEQ_END_CODE */
    
    data_stream->EOF_flag = 1;
    DBUG_RETURN(0);
  }
  
  num_read = num_read >> 2;
  
#if 0
  if (swap) {
#endif
    lmark = (unsigned int *) mark;

    for (i = 0; i < num_read; i++) {
      *lmark = htonl(*lmark);
      lmark++;
    }
#if 0
  }
#endif
  
  *buf_ptr = buf_start;
  *length_ptr = length + num_read;
  
  DBUG_RETURN(1);
}



#ifdef SYS_STREAM

/* 
  Here is the specialist.... 
  Code is adapted from our program demux....
  A bunch of this needs to be #ifdef ANALYSIS'ed
  define __SYSREAD_LOGGING_ON__ to get  an output file for debugging
  */

/* Brown - removed StreamID global variables */

/*
 *----------------------------------------------------------
 *
 *  read_sys
 *
 *      Parse out a packet of the system layer MPEG file.
 *
 *  Results:  Returns 0 if error or EOF
 *            Returns 1 if more data read (could be just one int)
 *
 *  Side Effects:  ReadPacket can change *bs_ptr to be a new buffer
 *                 buf_ptr will remain pointing at *length_ptr (at input)
 *                         into the buffer
 *                 *length_ptr will be changed to the new size
 *                 *max_length can be changed if a new buffer is alloc'd
 *
 *----------------------------------------------------------
 */
int read_sys(data_stream, start)
     DataStream *data_stream;
     unsigned int start;  
     /* start is either a start code or 0 to indicate continued parsing */
{
  unsigned int **bs_ptr=&data_stream->buf_start;
  int *max_length = &data_stream->max_buf_length;
  int *length_ptr=&data_stream->buf_length;
  unsigned int **buf_ptr=&data_stream->buffer;
  unsigned int startCode;
  int errorCode, PacketReply;
  unsigned char packetID;
  double systemClockTime;
  unsigned long muxRate;
  BOOLEAN match;
  DBUG_ENTER();
  
  if (!start) {
    errorCode = ReadStartCode(&startCode,data_stream);
    if (data_stream->EOF_flag) DBUG_RETURN(0);
    if (errorCode != 0) {
      fprintf(stderr, "Unable to read initial pack start code\n");
      DBUG_RETURN(0);
    }}
  else {
    errorCode = 0;
    startCode = start;
  }
  
  while (1) {
    match=FALSE;
    if (startCode == PACK_START_CODE) {
      match = TRUE;
      errorCode = ReadPackHeader( &systemClockTime, &muxRate, data_stream);
      if (errorCode != 0) {
        fprintf(stderr, "Error in reading pack header\n");
        DBUG_RETURN(0);
      }
      errorCode = ReadStartCode( &startCode, data_stream );
      if (errorCode != 0) {
        fprintf(stderr, "Error in reading start code\n");
        DBUG_RETURN(0);
      }
    }
    if (startCode == SYSTEM_HEADER_START_CODE) {
      match = TRUE;
      errorCode = ReadSystemHeader(data_stream);
      if (errorCode != 0) {
        fprintf(stderr, "Error in reading system header\n");
        DBUG_RETURN(0);
      }
      errorCode = ReadStartCode( &startCode, data_stream );
      if (errorCode != 0) {
        fprintf(stderr,"Error in reading start code after system header\n");
        DBUG_RETURN(0);
      }
    }
    packetID = startCode & 0xff;
    while (((startCode & PACKET_START_CODE_MASK) == PACKET_START_CODE_PREFIX) &&
	   (packetID >= 0xbc)) {
      match = TRUE;
      packetID = startCode & 0xff;
      PacketReply = ReadPacket(packetID, data_stream);
      switch (PacketReply) {
      case 2: 
        DBUG_RETURN(1);
      case 1: 
        DBUG_RETURN(0);
      default: /* do nothing */
        break;
      }
      errorCode = ReadStartCode( &startCode, data_stream );
      if (errorCode != 0) {
        fprintf(stderr,"Error in start code after packet\n");
        DBUG_RETURN(0);
      }
      if (startCode == PACK_START_CODE || startCode == ISO_11172_END_CODE) {
        break;
      }
    }
    
    if (startCode == ISO_11172_END_CODE) {
      match = TRUE;
      if (data_stream->Parse_done) {
	DBUG_RETURN(1);
      }
      ReadPacket(NOT_PACKET_ID, data_stream);
      data_stream->Parse_done = TRUE;
      DBUG_RETURN(1);
    }
    if (errorCode != 0)
      DBUG_RETURN(1);
    if (! match) {
      fprintf(stderr,"\nNo match found for start code %08x in system layer, skipping\n",startCode);
      startCode = find_start_code(data_stream->input);
      if (startCode == EOF) {
        data_stream->EOF_flag = 1;
        DBUG_RETURN(0);
      }
    }
  }
}


/*
 *-----------------------------------------------------------
 *
 *  ReadStartCode
 *
 *      Parses a start code out of the stream
 *
 *  Results/Side Effects:  Sets *startCode to the code, returns
 *     1 on error, 0 on success
 *
 *-----------------------------------------------------------
 */
int ReadStartCode(startCode, data_stream)
     unsigned int *startCode;
     DataStream *data_stream;
{
  int numRead;
  DBUG_ENTER();
  
  numRead = fread((unsigned char *)startCode, 1, 4, data_stream->input);
  *startCode = htonl(*startCode);
  
  if (numRead < 4) {
    data_stream->EOF_flag = 1;
    DBUG_RETURN(1);
  }

  if ((*startCode&0xfffffe00) != 0) {
    fprintf(stderr,"Problem with system layer parse, skipping to start code\n");
    *startCode = find_start_code(data_stream->input);
    if (*startCode == EOF) {
       data_stream->EOF_flag = TRUE;
      DBUG_RETURN(0);
    }
  }

  DBUG_RETURN(0);
}


/*
 *-----------------------------------------------------------
 *
 *  find_start_code
 *
 *      Parses a start code out of the stream by tossing bytes until it gets one
 *
 *  Results/Side Effects:  Parses bytes of the stream, returns code
 *                         Returns EOF in case of end of file
 *
 *-----------------------------------------------------------
 */
int find_start_code(input)
    FILE *input;
{
  DBUG_ENTER();
 NO_ZEROS:
  switch(fgetc(input)) {
  case 0:    goto ONE_ZERO;
  case EOF:  goto EOF_FOUND;
  default:   goto NO_ZEROS;
  }

 ONE_ZERO:
  switch(fgetc(input)) {
  case 0:    goto TWO_ZEROS;
  case EOF:  goto EOF_FOUND;
  default:   goto NO_ZEROS;
  }

 TWO_ZEROS:
  switch(fgetc(input)) {
  case 0x01:  goto CODE_FOUND;
  case 0x00:  goto TWO_ZEROS;
  case EOF:  goto EOF_FOUND;
  default:    goto NO_ZEROS;
  }

 CODE_FOUND:
  DBUG_RETURN(0x00000100+fgetc(input));

 EOF_FOUND:   /* received EOF */
  DBUG_RETURN(EOF);
}




/*
 *-----------------------------------------------------------------
 *
 *  ReadPackHeader
 *
 *      Parses out the PACK header
 *
 *  Returns: 1 on error, 0 on success
 *
 *-------------------------------------------------------------------
 */
int ReadPackHeader(systemClockTime,muxRate, data_stream)
     double *systemClockTime;
     unsigned long *muxRate;
     DataStream *data_stream;
{
  int numRead;
  unsigned char inputBuffer[PACK_HEADER_SIZE];
  unsigned long systemClockRef;
  unsigned char systemClockRefHiBit;
  int errorCode;
  DBUG_ENTER();
  
  numRead = fread(inputBuffer, 1, PACK_HEADER_SIZE, data_stream->input);
  if (numRead < PACK_HEADER_SIZE) {
    data_stream->EOF_flag = 1;
    DBUG_RETURN(1);
  }
  ReadTimeStamp(inputBuffer, &systemClockRefHiBit, &systemClockRef);
  errorCode = MakeFloatClockTime(systemClockRefHiBit, systemClockRef, 
				 systemClockTime);
  ReadRate(&inputBuffer[5], muxRate);
  *muxRate *= MUX_RATE_SCALE_FACTOR;
  DBUG_RETURN(0);
}


/*
 *------------------------------------------------------------------
 *
 *   ReadSystemHeader
 *
 *      Parse out the system header, setup out stream IDs for parsing packets
 *
 *   Results:  Returns 1 on error, 0 on success.
 *             Sets gAudioStreamID and gVideoStreamID
 *
 *------------------------------------------------------------------
 */
int ReadSystemHeader(data_stream)
   DataStream *data_stream;
{ 
  unsigned char *inputBuffer = NULL;
  int numRead;
  int pos;
  unsigned short headerSize;
  unsigned char streamID;
  DBUG_ENTER();
  
  numRead = fread((char *)&headerSize, 1, 2, data_stream->input); 
  headerSize = ntohs(headerSize);
  if (numRead != 2) {
    data_stream->EOF_flag = 1;
    DBUG_RETURN(1);
  }
  inputBuffer = (unsigned char *) malloc((unsigned int) headerSize+1);
  if (inputBuffer == NULL) {
    DBUG_RETURN(1);
  }
  inputBuffer[headerSize]=0;
  numRead = fread(inputBuffer, 1, headerSize, data_stream->input); 
  /* Brown - get rid of Ansi C complaints */
  if (numRead < (int) headerSize) {
    data_stream->EOF_flag = 1;
    DBUG_RETURN(1);
  }
  
  pos = 6;
  while ((inputBuffer[pos] & 0x80) == 0x80) {
    streamID = inputBuffer[pos];
    switch (streamID) {
    case STD_VIDEO_STREAM_ID: 
      break;
    case STD_AUDIO_STREAM_ID: 
      break;
    case RESERVED_STREAM_ID: 
      break;
    case PADDING_STREAM_ID: 
      break;
    case PRIVATE_STREAM_1_ID: 
      break;
    case PRIVATE_STREAM_2_ID: 
      break;
    default:
      if (streamID < MIN_STREAM_ID_ID) {
	DBUG_RETURN(1);
      }
      switch (streamID >> 4) {
      case 0xc:
      case 0xd:
	data_stream->gAudioStreamID = streamID;
	break;
      case 0xe:
	if ((data_stream->gVideoStreamID != 0) &&
	    (data_stream->gVideoStreamID!=streamID)) {
	  break;
	}
	data_stream->gVideoStreamID = streamID;
	break;
      case 0xf:
/*Brown - deglobalized gReservedStreamID */
	data_stream->gReservedStreamID = streamID;
	break;
      }
      break;
    }
    pos += 3;
  }
  if (inputBuffer != NULL)
    free(inputBuffer);
  DBUG_RETURN(0);
}


/*
 *-----------------------------------------------------------------
 *
 *  ReadPacket
 *
 *      Reads a single packet out of the stream, and puts it in the
 *      buffer if it is video.
 *
 *  Results:
 *      Changes the value of *length_ptr to be the new length (plus old)
 *      If the buffer is too small, can change *bs_ptr, *max_length, and 
 *      buf_ptr to be correct for a newly allocated buffer.
 *
 *  State:  
 *      The buffer is in ints, but the packets can be an arbitrary number
 *      of bytes, so leftover bytes are kept in the DataStream variable and
 *      are added on the next call.
 *
 *-----------------------------------------------------------------
 */   
#ifdef __STDC__
int ReadPacket(unsigned char packetID, DataStream *data_stream)
#else
int ReadPacket(packetID, data_stream)
     unsigned char packetID;
     DataStream *data_stream;
#endif

     /* Returns:
	0 - no error, but not video packet we want
	1 - error
	2 - got video packet into buffer
	*/
{   
  unsigned int **bs_ptr=&data_stream->buf_start;
  int *max_length = &data_stream->max_buf_length;
  int *length_ptr=&data_stream->buf_length;
  unsigned int **buf_ptr=&data_stream->buffer;
  int ioBytes;
  unsigned char nextByte;
  unsigned short packetLength;
  unsigned char *packetBuffer = NULL;
  int pos;
  int numStuffBytes = 0;
  unsigned int packetDataLength;
  int byte_length;
  unsigned char scratch[10];
  DBUG_ENTER();
  /* Leftovers from previous video packets */
  
  if (packetID == NOT_PACKET_ID) {
    /* Gross hack to handle unread bytes before end of stream */
    if (data_stream->num_left != 0) {
      /* Sigh, deal with previous leftovers */
      *(*buf_ptr+*length_ptr) = data_stream->leftover_bytes;
      *(*buf_ptr+*length_ptr+1) = ISO_11172_END_CODE;
      *length_ptr += 2;
    } else {
      *(*buf_ptr+*length_ptr) = ISO_11172_END_CODE;
      *length_ptr += 1;
    }
    DBUG_RETURN(1);
  } else if (packetID==KILL_BUFFER) {
    data_stream->num_left=0;
    data_stream->leftover_bytes=0;
    DBUG_RETURN(0);
  }
  
  ioBytes = fread(&packetLength, 1, 2, data_stream->input);
  packetLength = htons(packetLength);
  if (ioBytes < 2) {
    DBUG_RETURN(1);
  }
  if (packetID == data_stream->gAudioStreamID) {
  }
  else if (packetID == data_stream->gVideoStreamID) {
  }
  else {
    switch (packetID) {
    case PADDING_STREAM_ID:
      break;
    case RESERVED_STREAM_ID:
      break;
    case PRIVATE_STREAM_1_ID:
      break;
    case PRIVATE_STREAM_2_ID:
      break;
    default:
      fprintf(stderr, "\nUnknown packet type encountered. P'bly audio? (%x) at %d\n",
	      packetID,(int) ftell(data_stream->input));
    }
    if (packetID != data_stream->gVideoStreamID) {/* changed by jim */
      fseek(data_stream->input, packetLength, 1);
      DBUG_RETURN(0);
    }
  }

  fread(&nextByte,1,1,data_stream->input);
  pos = 0;
  while (nextByte & 0x80) {
    ++numStuffBytes;
    ++pos;
    fread(&nextByte,1,1,data_stream->input);
  }
  if ((nextByte >> 6) == 0x01) {
    pos += 2;
    fread(&nextByte,1,1,data_stream->input);
    fread(&nextByte,1,1,data_stream->input);
  } 
  if ((nextByte >> 4) == 0x02) {
    scratch[0] = nextByte;                      /* jim */
    fread(&scratch[1],1,4,data_stream->input);   /* jim */
    fread(&nextByte,1,1,data_stream->input);
    pos += 5;
  }
  else if ((nextByte >> 4) == 0x03) {
    scratch[0] = nextByte;                      /* jim */
    fread(&scratch[1],1,9,data_stream->input);   /* jim */
    fread(&nextByte,1,1,data_stream->input);
    pos += 10;
  } 
  else {
    fread(&nextByte,1,1,data_stream->input);
    pos += 1;
  }
  /* Read all the headers, now make room for packet */
  if (*bs_ptr + *max_length < *buf_ptr+ packetLength/4 + *length_ptr) {
     /* Brown - get rid of Ansi C complaints */
    if (*max_length - *length_ptr < (int) packetLength/4) {
      /* Buffer too small for a packet (plus whats there), 
	   * time to enlarge it! 
	   */
      unsigned int *old = *bs_ptr;
      *max_length = *length_ptr + packetLength/2;
      *bs_ptr=(unsigned int *)malloc(*max_length*4);
      if (*bs_ptr == NULL) {
        DBUG_RETURN(1);
      }
      memcpy((unsigned char *)*bs_ptr, *buf_ptr, (unsigned int) *length_ptr*4);
      free(old);
      *buf_ptr = *bs_ptr;
    } else {
      memcpy((unsigned char *)*bs_ptr, *buf_ptr, (unsigned int) *length_ptr*4);
      *buf_ptr = *bs_ptr;
    }}
  byte_length = *length_ptr*4;
  if (data_stream->num_left != 0) {
    /* Sigh, deal with previous leftovers */
    byte_length += data_stream->num_left;
    *(*buf_ptr+*length_ptr) = data_stream->leftover_bytes;
  }
  packetBuffer=((unsigned char *)*buf_ptr)+byte_length;
  packetDataLength = packetLength - pos;
  *packetBuffer++ = nextByte;
/* Brown - deglobalize gVideoStreamID */
  if (packetID == data_stream->gVideoStreamID) {
    ioBytes = fread(packetBuffer, 1, packetDataLength-1, data_stream->input);
    if (ioBytes != packetDataLength-1) {
      data_stream->EOF_flag = 1;
      DBUG_RETURN(1);
    }
    if (1 != ntohl(1)) {
      unsigned int *mark = *buf_ptr+*length_ptr;
      int i;
      
      for (i=0; i < ((packetDataLength+
			 data_stream->num_left)&0xfffffffc); i+=4) {
        *mark=ntohl(*mark);
        mark++;
      }
    }
    byte_length = byte_length + packetDataLength;
    data_stream->num_left = byte_length % 4;
    *length_ptr = byte_length / 4;
    data_stream->leftover_bytes = *(*buf_ptr + *length_ptr);
    DBUG_RETURN(2);
  }
  else if (packetID == data_stream->gAudioStreamID) { 
    packetBuffer = (unsigned char *)(*buf_ptr + *length_ptr + 1);
    fread(packetBuffer, 1, packetDataLength - 1, data_stream->input);
  }
  else /* Donno what it is, just nuke it */ {
    /* This code should be unreachable */
    packetBuffer = (unsigned char *)(*buf_ptr + *length_ptr + 1);
    fread(packetBuffer, 1, packetDataLength - 1, data_stream->input);
  }
  DBUG_RETURN(0); 
}


/*
 * The remaining procedures are formatting utility procedures.
 */
void ReadTimeStamp(inputBuffer,hiBit,low4Bytes)
     unsigned char *inputBuffer, *hiBit;
     unsigned long *low4Bytes;
{
  DBUG_ENTER();
  *hiBit = ((unsigned long)inputBuffer[0] >> 3) & 0x01;
  *low4Bytes = (((unsigned long)inputBuffer[0] >> 1) & 0x03) << 30; 
  *low4Bytes |= (unsigned long)inputBuffer[1] << 22; 
  *low4Bytes |= ((unsigned long)inputBuffer[2] >> 1) << 15; 
  *low4Bytes |= (unsigned long)inputBuffer[3] << 7; 
  *low4Bytes |= ((unsigned long)inputBuffer[4]) >> 1; 
  DBUG_VOID_RETURN();
}

void ReadSTD(inputBuffer,stdBufferScale,stdBufferSize) 
unsigned char *inputBuffer;
unsigned char *stdBufferScale;
unsigned long *stdBufferSize;
{
  DBUG_ENTER();
  /* Brown - get rid of ANSI C complaints */
  *stdBufferScale = ((int)(inputBuffer[0] & 0x20) >> 5); 
  *stdBufferSize = ((unsigned long)inputBuffer[0] & 0x1f) << 8;
  *stdBufferSize |= (unsigned long)inputBuffer[1];
  DBUG_VOID_RETURN();
}


void ReadRate(inputBuffer,rate)
     unsigned char *inputBuffer;
     unsigned long *rate;
{
  DBUG_ENTER();
  *rate = (inputBuffer[0] & 0x7f) << 15;
  *rate |= inputBuffer[1] << 7;
  /* Brown - get rid of ANSI C complaints */
  *rate |= (int) (inputBuffer[2] & 0xfe) >> 1;
  DBUG_VOID_RETURN();
}

#define FLOAT_0x10000 (double)((unsigned long)1 << 16)

#ifdef __STDC__
int MakeFloatClockTime(unsigned char hiBit, unsigned long low4Bytes,
		       double * floatClockTime)
#else
int MakeFloatClockTime(hiBit,low4Bytes,floatClockTime)
     unsigned char hiBit;
     unsigned long low4Bytes;
     double *floatClockTime;
#endif
{
  DBUG_ENTER();
  if (hiBit != 0 && hiBit != 1) {
    *floatClockTime = 0.0;
    DBUG_RETURN(1);
  }
  *floatClockTime 
    = (double)hiBit*FLOAT_0x10000*FLOAT_0x10000 + (double)low4Bytes;
  *floatClockTime /= (double)STD_SYSTEM_CLOCK_FREQ;
  DBUG_RETURN(0);
}

#endif /* SYS_STREAM */
