/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/*
 * util.c --
 *
 *      Miscellaneous utility procedures.
 *
 */

#include "global.h"

#include <stdlib.h>
#include <assert.h>

#include "readfile.h"
#include "util.h"
#include "video.h"


/*
 *--------------------------------------------------------------
 *
 * next_start_code --
 *
 *	Parses off bitstream until start code reached. When done
 *      next 4 bytes of bitstream will be start code. Bit offset
 *      reset to 0.
 *
 * Results:
 *	Status code.
 *
 * Side effects:
 *	Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

int next_start_code(data_stream)
   DataStream *data_stream;
{
  int byteoff;
  unsigned int data;
  DBUG_ENTER();

  /* If no current stream, return error. */

  if (data_stream == NULL)
    DBUG_RETURN(NO_VID_STREAM);

  /* If insufficient buffer length, correct underflow. */

  if (data_stream->buf_length < 4) {
    correct_underflow(data_stream);
  }

  /* If bit offset not zero, reset and advance buffer pointer. */

  byteoff = data_stream->bit_offset & 7;

  if (byteoff != 0) {
    flush_bits((8-byteoff));
  }

  /* While buffer has data ... */

  while(data_stream->buf_length > 0) {

    /* If insufficient data exists, correct underflow. */
    
    if (data_stream->buf_length < 4) {
      correct_underflow(data_stream);
    }

    /* Return success if start code up next... */
    
    show_bits24(data);
    if (data == 0x000001)
      DBUG_RETURN(OK);

    /* Otherwise, advance a byte and try again. */
    
    flush_bits(8);
  }

  /* Return underflow error. */
  DBUG_RETURN(STREAM_UNDERFLOW);
}


/*
 *--------------------------------------------------------------
 *
 * get_ext_data --
 *
 *	Assumes that bit stream is at begining of extension
 *      data. Parses off extension data into dynamically 
 *      allocated space until start code is hit. 
 *
 * Results:
 *	Pointer to dynamically allocated memory containing
 *      extension data.
 *
 * Side effects:
 *	Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

char *get_ext_data (data_stream)
   DataStream *data_stream;
{
  unsigned int size, marker;
  char *dataPtr;
  unsigned int data;
  DBUG_ENTER();

  /* Set initial ext data buffer size. */

  size = EXT_BUF_SIZE;

  /* Allocate ext data buffer. */

  dataPtr = (char *) malloc(size);

  /* Initialize marker to keep place in ext data buffer. */

  marker = 0;

  /* While next data is not start code... */
  while (!next_bits(24, 0x000001, data_stream)) {

    /* Get next byte of ext data. */

    get_bits8(data);

    /* Put ext data into ext data buffer. Advance marker. */

    dataPtr[marker] = (char) data;
    marker++;

    /* If end of ext data buffer reached, resize data buffer. */

    if (marker == size) {
      size += EXT_BUF_SIZE;
      dataPtr = (char *) realloc(dataPtr, size);
    }
  }

  /* Realloc data buffer to free any extra space. */

  dataPtr = (char *) realloc(dataPtr, marker);

  /* Return pointer to ext data buffer. */
  DBUG_RETURN(dataPtr);
}


/*
 *--------------------------------------------------------------
 *
 * get_extra_bit_info --
 *
 *	Parses off extra bit info stream into dynamically 
 *      allocated memory. Extra bit info is indicated by
 *      a flag bit set to 1, followed by 8 bits of data.
 *      This continues until the flag bit is zero. Assumes
 *      that bit stream set to first flag bit in extra
 *      bit info stream.
 *
 * Results:
 *	Pointer to dynamically allocated memory with extra
 *      bit info in it. Flag bits are NOT included.
 *
 * Side effects:
 *	Bit stream irreversibly parsed.
 *
 *--------------------------------------------------------------
 */

char *get_extra_bit_info(data_stream)
DataStream *data_stream;
{
  unsigned int size, marker;
  char *dataPtr;
  unsigned int data;
  DBUG_ENTER();

  /* Get first flag bit. */
  get_bits1(data);

  /* If flag is false, return NULL pointer (i.e. no extra bit info). */

  if (!data) DBUG_RETURN(NULL);

  /* Initialize size of extra bit info buffer and allocate. */

  size = EXT_BUF_SIZE;
  dataPtr = (char *) malloc(size);

  /* Reset marker to hold place in buffer. */

  marker = 0;

  /* While flag bit is true. */

  while (data) {

    /* Get next 8 bits of data. */
    get_bits8(data);

    /* Place in extra bit info buffer. */

    dataPtr[marker] = (char) data;
    marker++;

    /* If buffer is full, reallocate. */

    if (marker == size) {
      size += EXT_BUF_SIZE;
      dataPtr = (char *) realloc(dataPtr, size);
    }

    /* Get next flag bit. */
    get_bits1(data);
  }

  /* Reallocate buffer to free extra space. */

  dataPtr = (char *) realloc(dataPtr, marker);

  /* Return pointer to extra bit info buffer. */
  DBUG_RETURN(dataPtr);
}
