#include "streamer-private.h"

void _StreamBuffer_init() {
  DEBUG("in _StreamBuffer_init()\n");
}

void StreamBuffer_init(StreamBuffer *buf) {
  buf->mutex = g_mutex_new();
  buf->refcount = 0;
  buf->typemetasize = 0;
  buf->typemeta = NULL;
  buf->metasize = 0;
  buf->meta = NULL;
  buf->datasize = 0;
  buf->dataused = 0;
  buf->data = NULL;
  buf->copyfunc = NULL;
  buf->freefunc = NULL;
}

void StreamBuffer_poolinit(StreamBufferPool *pool,StreamBuffer *buf) {
  gpointer ptr = (gpointer)buf;

  StreamBuffer_init(buf);

  ptr += sizeof(StreamBuffer);
  if (pool->typemetasize) {
//    g_print("allocating %d bytes to typemeta\n",pool->typemetasize);
    buf->typemetasize = pool->typemetasize;
    buf->typemeta = ptr;
    ptr += buf->typemetasize;
  }
  if (pool->metasize) {
//    g_print("allocating %d bytes to meta\n",pool->metasize);
    buf->metasize = pool->metasize;
    buf->meta = ptr;
    ptr += buf->metasize;
  }
  if (pool->datasize) {
//    g_print("allocating %d bytes to data\n",pool->datasize);
    buf->datasize = pool->datasize;
    buf->data = ptr;
  }
}

StreamBuffer *StreamBuffer_create(StreamBufferPool *pool) {
  StreamBuffer *buf;
  StreamBufferChunk *chunk;
  GSList *chunklist;
  int i;

  /* insanity check */
  if (!pool) {
    g_warning("StreamBuffer_create: pool is NULL!\n");
    return(NULL);
  }

  /* first lock the pool */
  g_mutex_lock(pool->mutex);

  /* first check to see if a resize migration is occuring */
  if (pool->newchunk) {
    /* see if there's space in the aux pool */
    if (pool->newused < pool->newsize) {
      buf = g_chunk_new(StreamBuffer,pool->newchunk);
      buf->pool = pool;
      buf->poolchunk = pool->newchunk;
      StreamBuffer_poolinit(pool,buf);
      pool->newused++;
      pool->buffers++;
      g_mutex_unlock(pool->mutex);
      return(buf);
    } else {
      /* we have an idiot at the helm mis-guessing the resize... */
      DEBUG("StreamBuffer_create: pool resize in progress, aux full!\n");
    }
  } else {
    /* see if there's space in the main pool */
    if (pool->used < pool->size) {
      buf = g_chunk_new(StreamBuffer,pool->chunk);
      buf->pool = pool;
      buf->poolchunk = pool->chunk;
      StreamBuffer_poolinit(pool,buf);
      pool->used++;
      pool->buffers++;
      g_mutex_unlock(pool->mutex);
      return(buf);
    }
  }

  /* from this point on, we're dealing with overflow buffers in ->extra */
  DEBUG("StreamBuffer_create: too many buffers, overflowing (nonfatal)\n");

  /* for each existing extra chunk, see if there's space */
  chunklist = pool->extra;
  i = 0;
  while (chunklist) {
    chunk = chunklist->data;
    if (chunk->used < chunk->size) {
      buf = g_chunk_new(StreamBuffer,chunk->chunk);
      buf->pool = pool;
      buf->poolchunk = chunk->chunk;
      StreamBuffer_poolinit(pool,buf);
      chunk->used++;
      pool->buffers++;
      g_mutex_unlock(pool->mutex);
      return(buf);
    }
    chunklist = g_slist_next(chunklist);
  }

  /* if we get to this point, we get to allocate an 'extra' pool */
  /* first create the extra chunk */
  DEBUG("StreamBuffer_create: allocating extra buffer pool\n");
  chunk = (StreamBufferChunk *)g_malloc(sizeof(StreamBufferChunk));
  chunk->chunk = g_mem_chunk_create(StreamBuffer,pool->size,
                                    G_ALLOC_AND_FREE);
  chunk->size = pool->size;
  chunk->used = 0;

  /* add it to the list of extra chunks */
  DEBUG("adding %p to list %p\n",chunk,pool->extra);
  pool->extra = g_slist_prepend(pool->extra,chunk);
  DEBUG("StreamBuffer_create: allocated extra buffer pool\n");

  /* now populate it with a buffer */
  buf = g_chunk_new(StreamBuffer,chunk->chunk);
  buf->pool = pool;
  buf->poolchunk = chunk->chunk;
  StreamBuffer_poolinit(pool,buf);
  chunk->used++;
  pool->buffers++;
  g_mutex_unlock(pool->mutex);
  return(buf);
}

void StreamBuffer_free(StreamBuffer *buf) {
  StreamBufferPool *pool;
  StreamBufferChunk *chunk;
  GSList *chunklist;

  /* insanity check */
  if (!buf) {
    g_warning("StreamBuffer_free: buffer is NULL\n");
    return;
  }

  /* we'll lock the buffer, providing a warning in case it is already */
  if (g_mutex_trylock(buf->mutex) == FALSE) {
    g_warning("StreamBuffer_free: buffer still locked, waiting\n");
    g_mutex_lock(buf->mutex);
  }

  /* verify that it's safe to free this think, i.e. the refcount is 0 */
  /* FIXME: there's probably a better way to handle this */
  if (buf->refcount) {
    g_error("StreamBuffer_free: buffer is still in use!\n");
    exit(1);
  }

  DEBUG("StreamBuffer_free: buffer is unencumbered, freeing\n");

  /* initial task is to have the buffer free its own data structures */
  if (buf->freefunc)
    (buf->freefunc)(buf);
  else {
    /* we have to do it the brute-force way */
    DEBUG("StreamBuffer_free: no free function in buffer!\n");
    if (!buf->pool->typemetasize)
      if (buf->typemeta) g_free(buf->typemeta),g_print("free typemeta\n");
    if (!buf->pool->metasize)
      if (buf->meta) g_free(buf->meta),g_print("free meta\n");
    if (!buf->pool->datasize)
      if (buf->data) g_free(buf->data),g_print("free data\n");
  }

  /* find the associated pool and lock it */
  pool = buf->pool;
  DEBUG("attempting to lock mutex\n");
  g_mutex_lock(pool->mutex);
  DEBUG("locked mutex\n");

  /* first see if it's in the main pool */
  if (buf->poolchunk == pool->chunk) {
    g_chunk_free(buf,pool->chunk);
    pool->used--;
    pool->buffers--;

    /* take care of slow resize case */
    if ((pool->used == 0) && (pool->newchunk != NULL)) {
      DEBUG("StreamBuffer_free: resize complete, new size %d\n",
            pool->newsize);

      /* move the new chunk pool into place */
      pool->chunk = pool->newchunk;
      pool->size = pool->newsize;
      pool->used = pool->newused;

      /* clean up the new chunk pool */
      pool->newchunk = NULL;
      pool->newsize = 0;
      pool->newused = 0;
    }

    g_mutex_unlock(pool->mutex);
    return;

  /* now check to see if it's in the new pool */
  } else if (buf->poolchunk == pool->newchunk) {
    g_chunk_free(buf,pool->newchunk);
    pool->newused--;
    pool->buffers--;
    g_mutex_unlock(pool->mutex);
    return;

  /* only thing left are the extra pools */
  } else {
    if (pool->extra == NULL) {
      g_warning("StreamBuffer_free: buffer doesn't seem to exist\n");
      g_mutex_unlock(pool->mutex);
      return;
    }

    chunklist = pool->extra;
    while (chunklist) {
      chunk = chunklist->data;
      if (buf->poolchunk == chunk->chunk) {
        g_chunk_free(buf,chunk->chunk);
        chunk->used--;
        pool->buffers--;

        /* see if we just cleaned out an extra chunk */
        if (chunk->used == 0) {
          DEBUG("StreamBuffer_free: cleaned out extra chunk, freeing\n");
          g_mem_chunk_destroy(chunk->chunk);
          pool->extra = g_slist_remove(pool->extra,chunk);
          if (pool->extra == NULL)
            DEBUG("StreamBuffer_free: cleaned out all extra chunks!\n");
        }
        g_mutex_unlock(pool->mutex);
        return;
      }
      chunklist = g_slist_next(chunklist);
    }
  } 

  /* if we haven't found it by now, there's something funky */
  g_warning("StreamBuffer_free: buffer doesn't seem to exist\n");
  g_mutex_unlock(pool->mutex);
}

gboolean StreamBuffer_nullfree(StreamBuffer *buf) {
  DEBUG("StreamBuffer_nullfree: returning TRUE\n");
  return(TRUE);
}
